#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock, patch

import pytest

from airflow.models.connection import Connection
from airflow.providers.common.compat.openlineage.facet import (
    ColumnLineageDatasetFacet,
    Dataset,
    Fields,
    InputField,
    SQLJobFacet,
)
from airflow.providers.common.sql.hooks.handlers import fetch_all_handler
from airflow.providers.common.sql.operators.sql import SQLExecuteQueryOperator
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook


class MockRow:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __eq__(self, other):
        return isinstance(other, MockRow) and self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(self.__dict__)

    def __repr__(self):
        return f"MockRow({self.__dict__})"


DATE = "2017-04-20"
TASK_ID = "databricks-sql-operator"
DEFAULT_CONN_ID = "snowflake_default"


@pytest.mark.parametrize(
    ("sql", "return_last", "split_statement", "hook_results", "hook_descriptions", "expected_results"),
    [
        pytest.param(
            "select * from dummy",
            True,
            True,
            [MockRow(id=1, value="value1"), MockRow(id=2, value="value2")],
            [[("id",), ("value",)]],
            ([MockRow(id=1, value="value1"), MockRow(id=2, value="value2")]),
            id="Scalar: Single SQL statement, return_last, split statement",
        ),
        pytest.param(
            "select * from dummy;select * from dummy2",
            True,
            True,
            [MockRow(id=1, value="value1"), MockRow(id=2, value="value2")],
            [[("id",), ("value",)]],
            ([MockRow(id=1, value="value1"), MockRow(id=2, value="value2")]),
            id="Scalar: Multiple SQL statements, return_last, split statement",
        ),
        pytest.param(
            "select * from dummy",
            False,
            False,
            [MockRow(id=1, value="value1"), MockRow(id=2, value="value2")],
            [[("id",), ("value",)]],
            ([MockRow(id=1, value="value1"), MockRow(id=2, value="value2")]),
            id="Scalar: Single SQL statements, no return_last (doesn't matter), no split statement",
        ),
        pytest.param(
            "select * from dummy",
            True,
            False,
            [MockRow(id=1, value="value1"), MockRow(id=2, value="value2")],
            [[("id",), ("value",)]],
            ([MockRow(id=1, value="value1"), MockRow(id=2, value="value2")]),
            id="Scalar: Single SQL statements, return_last (doesn't matter), no split statement",
        ),
        pytest.param(
            ["select * from dummy"],
            False,
            False,
            [[MockRow(id=1, value="value1"), MockRow(id=2, value="value2")]],
            [[("id",), ("value",)]],
            [([MockRow(id=1, value="value1"), MockRow(id=2, value="value2")])],
            id="Non-Scalar: Single SQL statements in list, no return_last, no split statement",
        ),
        pytest.param(
            ["select * from dummy", "select * from dummy2"],
            False,
            False,
            [
                [MockRow(id=1, value="value1"), MockRow(id=2, value="value2")],
                [MockRow(id2=1, value2="value1"), MockRow(id2=2, value2="value2")],
            ],
            [[("id",), ("value",)], [("id2",), ("value2",)]],
            [
                ([MockRow(id=1, value="value1"), MockRow(id=2, value="value2")]),
                ([MockRow(id2=1, value2="value1"), MockRow(id2=2, value2="value2")]),
            ],
            id="Non-Scalar: Multiple SQL statements in list, no return_last (no matter), no split statement",
        ),
        pytest.param(
            ["select * from dummy", "select * from dummy2"],
            True,
            False,
            [
                [MockRow(id=1, value="value1"), MockRow(id=2, value="value2")],
                [MockRow(id2=1, value2="value1"), MockRow(id2=2, value2="value2")],
            ],
            [[("id",), ("value",)], [("id2",), ("value2",)]],
            [
                ([MockRow(id=1, value="value1"), MockRow(id=2, value="value2")]),
                ([MockRow(id2=1, value2="value1"), MockRow(id2=2, value2="value2")]),
            ],
            id="Non-Scalar: Multiple SQL statements in list, return_last (no matter), no split statement",
        ),
    ],
)
def test_exec_success(sql, return_last, split_statement, hook_results, hook_descriptions, expected_results):
    """
    Test the execute function in case where SQL query was successful.
    """
    with patch("airflow.providers.common.sql.operators.sql.BaseSQLOperator.get_db_hook") as get_db_hook_mock:
        op = SQLExecuteQueryOperator(
            task_id=TASK_ID,
            sql=sql,
            do_xcom_push=True,
            return_last=return_last,
            split_statements=split_statement,
            conn_id="snowflake_default",
        )
        dbapi_hook = MagicMock()
        get_db_hook_mock.return_value = dbapi_hook
        dbapi_hook.run.return_value = hook_results
        dbapi_hook.descriptions = hook_descriptions

        execute_results = op.execute(None)

        assert execute_results == expected_results
        dbapi_hook.run.assert_called_once_with(
            sql=sql,
            parameters=None,
            handler=fetch_all_handler,
            autocommit=False,
            return_last=return_last,
            split_statements=split_statement,
        )


@mock.patch("airflow.providers.openlineage.utils.utils.should_use_external_connection")
def test_execute_openlineage_events(should_use_external_connection):
    should_use_external_connection.return_value = False
    DB_NAME = "DATABASE"
    DB_SCHEMA_NAME = "PUBLIC"

    ANOTHER_DB_NAME = "ANOTHER_DB"
    ANOTHER_DB_SCHEMA = "ANOTHER_SCHEMA"

    class SnowflakeHookForTests(SnowflakeHook):
        get_conn = MagicMock(name="conn")
        get_connection = MagicMock()

    dbapi_hook = SnowflakeHookForTests()

    class SnowflakeOperatorForTest(SQLExecuteQueryOperator):
        def get_db_hook(self):
            return dbapi_hook

    sql = (
        "INSERT INTO Test_table\n"
        "SELECT t1.*, t2.additional_constant FROM ANOTHER_DB.ANOTHER_SCHEMA.popular_orders_day_of_week t1\n"
        "JOIN little_table t2 ON t1.order_day_of_week = t2.order_day_of_week;\n"
        "FORGOT TO COMMENT"
    )

    op = SnowflakeOperatorForTest(task_id="snowflake-operator", sql=sql)
    rows = [
        [
            (
                ANOTHER_DB_SCHEMA,
                "POPULAR_ORDERS_DAY_OF_WEEK",
                "ORDER_DAY_OF_WEEK",
                1,
                "TEXT",
                ANOTHER_DB_NAME,
            ),
            (
                ANOTHER_DB_SCHEMA,
                "POPULAR_ORDERS_DAY_OF_WEEK",
                "ORDER_PLACED_ON",
                2,
                "TIMESTAMP_NTZ",
                ANOTHER_DB_NAME,
            ),
            (ANOTHER_DB_SCHEMA, "POPULAR_ORDERS_DAY_OF_WEEK", "ORDERS_PLACED", 3, "NUMBER", ANOTHER_DB_NAME),
            (DB_SCHEMA_NAME, "LITTLE_TABLE", "ORDER_DAY_OF_WEEK", 1, "TEXT", DB_NAME),
            (DB_SCHEMA_NAME, "LITTLE_TABLE", "ADDITIONAL_CONSTANT", 2, "TEXT", DB_NAME),
        ],
        [
            (DB_SCHEMA_NAME, "TEST_TABLE", "ORDER_DAY_OF_WEEK", 1, "TEXT", DB_NAME),
            (DB_SCHEMA_NAME, "TEST_TABLE", "ORDER_PLACED_ON", 2, "TIMESTAMP_NTZ", DB_NAME),
            (DB_SCHEMA_NAME, "TEST_TABLE", "ORDERS_PLACED", 3, "NUMBER", DB_NAME),
            (DB_SCHEMA_NAME, "TEST_TABLE", "ADDITIONAL_CONSTANT", 4, "TEXT", DB_NAME),
        ],
    ]
    dbapi_hook.get_connection.return_value = Connection(
        conn_id="snowflake_default",
        conn_type="snowflake",
        schema="PUBLIC",
        extra={
            "account": "test_account",
            "region": "us-east",
            "warehouse": "snow-warehouse",
            "database": DB_NAME,
        },
    )
    dbapi_hook.get_conn.return_value.cursor.return_value.fetchall.side_effect = rows

    lineage = op.get_openlineage_facets_on_start()
    # Not calling Snowflake
    assert dbapi_hook.get_conn.return_value.cursor.return_value.execute.mock_calls == []

    assert lineage.inputs == [
        Dataset(
            namespace="snowflake://test_account.us-east.aws",
            name=f"{DB_NAME}.{DB_SCHEMA_NAME}.LITTLE_TABLE",
        ),
        Dataset(
            namespace="snowflake://test_account.us-east.aws",
            name=f"{ANOTHER_DB_NAME}.{ANOTHER_DB_SCHEMA}.POPULAR_ORDERS_DAY_OF_WEEK",
        ),
    ]
    assert lineage.outputs == [
        Dataset(
            namespace="snowflake://test_account.us-east.aws",
            name=f"{DB_NAME}.{DB_SCHEMA_NAME}.TEST_TABLE",
            facets={
                "columnLineage": ColumnLineageDatasetFacet(
                    fields={
                        "additional_constant": Fields(
                            inputFields=[
                                InputField(
                                    namespace="snowflake://test_account.us-east.aws",
                                    name="DATABASE.PUBLIC.little_table",
                                    field="additional_constant",
                                )
                            ],
                            transformationDescription="",
                            transformationType="",
                        )
                    }
                ),
            },
        )
    ]

    assert lineage.job_facets == {"sql": SQLJobFacet(query=sql)}

    assert lineage.run_facets["extractionError"].failedTasks == 1



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

pytest.importorskip("snowflake-snowpark-python")

from airflow.providers.snowflake.utils.snowpark import inject_session_into_op_kwargs


@pytest.mark.parametrize(
    ("func", "expected_injected"),
    [
        (lambda x: x, False),
        (lambda: 1, False),
        (lambda session: 1, True),
        (lambda session, x: x, True),
        (lambda x, session: 2 * x, True),
    ],
)
def test_inject_session_into_op_kwargs(func, expected_injected):
    result = inject_session_into_op_kwargs(func, {}, None)
    assert ("session" in result) == expected_injected



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# NOTE! THIS FILE IS AUTOMATICALLY GENERATED AND WILL BE OVERWRITTEN!
#
# IF YOU WANT TO MODIFY THIS FILE, YOU SHOULD MODIFY THE TEMPLATE
# `get_provider_info_TEMPLATE.py.jinja2` IN the `dev/breeze/src/airflow_breeze/templates` DIRECTORY


def get_provider_info():
    return {
        "package-name": "apache-airflow-providers-snowflake",
        "name": "Snowflake",
        "description": "`Snowflake <https://www.snowflake.com/>`__\n",
        "integrations": [
            {
                "integration-name": "Snowflake",
                "external-doc-url": "https://snowflake.com/",
                "how-to-guide": [
                    "/docs/apache-airflow-providers-snowflake/operators/snowflake.rst",
                    "/docs/apache-airflow-providers-snowflake/operators/snowpark.rst",
                ],
                "logo": "/docs/integration-logos/Snowflake.png",
                "tags": ["service"],
            }
        ],
        "operators": [
            {
                "integration-name": "Snowflake",
                "python-modules": [
                    "airflow.providers.snowflake.operators.snowflake",
                    "airflow.providers.snowflake.operators.snowpark",
                ],
            }
        ],
        "task-decorators": [
            {
                "class-name": "airflow.providers.snowflake.decorators.snowpark.snowpark_task",
                "name": "snowpark",
            }
        ],
        "hooks": [
            {
                "integration-name": "Snowflake",
                "python-modules": [
                    "airflow.providers.snowflake.hooks.snowflake",
                    "airflow.providers.snowflake.hooks.snowflake_sql_api",
                ],
            }
        ],
        "transfers": [
            {
                "source-integration-name": "Amazon Simple Storage Service (S3)",
                "target-integration-name": "Snowflake",
                "python-module": "airflow.providers.snowflake.transfers.copy_into_snowflake",
                "how-to-guide": "/docs/apache-airflow-providers-snowflake/operators/copy_into_snowflake.rst",
            },
            {
                "source-integration-name": "Google Cloud Storage (GCS)",
                "target-integration-name": "Snowflake",
                "python-module": "airflow.providers.snowflake.transfers.copy_into_snowflake",
                "how-to-guide": "/docs/apache-airflow-providers-snowflake/operators/copy_into_snowflake.rst",
            },
            {
                "source-integration-name": "Microsoft Azure Blob Storage",
                "target-integration-name": "Snowflake",
                "python-module": "airflow.providers.snowflake.transfers.copy_into_snowflake",
                "how-to-guide": "/docs/apache-airflow-providers-snowflake/operators/copy_into_snowflake.rst",
            },
        ],
        "connection-types": [
            {
                "hook-class-name": "airflow.providers.snowflake.hooks.snowflake.SnowflakeHook",
                "connection-type": "snowflake",
            }
        ],
        "triggers": [
            {
                "integration-name": "Snowflake",
                "python-modules": ["airflow.providers.snowflake.triggers.snowflake_trigger"],
            }
        ],
        "config": {
            "snowflake": {
                "description": "Configuration for Snowflake hooks and operators.\n",
                "options": {
                    "azure_oauth_scope": {
                        "description": "The scope to use while retrieving OAuth token for Snowflake from Azure Entra authentication.\n",
                        "version_added": "6.6.0",
                        "type": "string",
                        "example": None,
                        "default": "api://snowflake_oauth_server/.default",
                    }
                },
            }
        },
    }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import base64
import time
import uuid
import warnings
from datetime import timedelta
from pathlib import Path
from typing import Any

import aiohttp
import requests
from aiohttp import ClientConnectionError, ClientResponseError
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from requests.exceptions import ConnectionError, HTTPError, Timeout
from tenacity import (
    AsyncRetrying,
    Retrying,
    before_sleep_log,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential,
)

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook
from airflow.providers.snowflake.utils.sql_api_generate_jwt import JWTGenerator


class SnowflakeSqlApiHook(SnowflakeHook):
    """
    A client to interact with Snowflake using SQL API and submit multiple SQL statements in a single request.

    In combination with aiohttp, make post request to submit SQL statements for execution,
    poll to check the status of the execution of a statement. Fetch query results asynchronously.

    This hook requires the snowflake_conn_id connection. This hooks mainly uses account, schema, database,
    warehouse, and an authentication mechanism from one of below:
    1. JWT Token generated from private_key_file or private_key_content. Other inputs can be defined in the connection or hook instantiation.
    2. OAuth Token generated from the refresh_token, client_id and client_secret specified in the connection

    :param snowflake_conn_id: Reference to
        :ref:`Snowflake connection id<howto/connection:snowflake>`
    :param account: snowflake account name
    :param authenticator: authenticator for Snowflake.
        'snowflake' (default) to use the internal Snowflake authenticator
        'externalbrowser' to authenticate using your web browser and
        Okta, ADFS or any other SAML 2.0-compliant identify provider
        (IdP) that has been defined for your account
        'https://<your_okta_account_name>.okta.com' to authenticate
        through native Okta.
    :param warehouse: name of snowflake warehouse
    :param database: name of snowflake database
    :param region: name of snowflake region
    :param role: name of snowflake role
    :param schema: name of snowflake schema
    :param session_parameters: You can set session-level parameters at
        the time you connect to Snowflake
    :param token_life_time: lifetime of the JWT Token in timedelta
    :param token_renewal_delta: Renewal time of the JWT Token in timedelta
    :param deferrable: Run operator in the deferrable mode.
    :param api_retry_args: An optional dictionary with arguments passed to ``tenacity.Retrying`` & ``tenacity.AsyncRetrying`` classes.
    """

    LIFETIME = timedelta(minutes=59)  # The tokens will have a 59 minute lifetime
    RENEWAL_DELTA = timedelta(minutes=54)  # Tokens will be renewed after 54 minutes

    def __init__(
        self,
        snowflake_conn_id: str,
        token_life_time: timedelta = LIFETIME,
        token_renewal_delta: timedelta = RENEWAL_DELTA,
        api_retry_args: dict[Any, Any] | None = None,  # Optional retry arguments passed to tenacity.retry
        *args: Any,
        **kwargs: Any,
    ):
        self.snowflake_conn_id = snowflake_conn_id
        self.token_life_time = token_life_time
        self.token_renewal_delta = token_renewal_delta

        super().__init__(snowflake_conn_id, *args, **kwargs)
        self.private_key: Any = None

        self.retry_config = {
            "retry": retry_if_exception(self._should_retry_on_error),
            "wait": wait_exponential(multiplier=1, min=1, max=60),
            "stop": stop_after_attempt(5),
            "before_sleep": before_sleep_log(self.log, log_level=20),  # type: ignore[arg-type]
            "reraise": True,
        }
        if api_retry_args:
            self.retry_config.update(api_retry_args)

    def get_private_key(self) -> None:
        """Get the private key from snowflake connection."""
        conn = self.get_connection(self.snowflake_conn_id)

        # If private_key_file is specified in the extra json, load the contents of the file as a private key.
        # If private_key_content is specified in the extra json, use it as a private key.
        # As a next step, specify this private key in the connection configuration.
        # The connection password then becomes the passphrase for the private key.
        # If your private key is not encrypted (not recommended), then leave the password empty.

        private_key_file = conn.extra_dejson.get(
            "extra__snowflake__private_key_file"
        ) or conn.extra_dejson.get("private_key_file")
        private_key_content = conn.extra_dejson.get(
            "extra__snowflake__private_key_content"
        ) or conn.extra_dejson.get("private_key_content")

        private_key_pem = None
        if private_key_content and private_key_file:
            raise AirflowException(
                "The private_key_file and private_key_content extra fields are mutually exclusive. "
                "Please remove one."
            )
        if private_key_file:
            private_key_pem = Path(private_key_file).read_bytes()
        elif private_key_content:
            private_key_pem = base64.b64decode(private_key_content)

        if private_key_pem:
            passphrase = None
            if conn.password:
                passphrase = conn.password.strip().encode()

            self.private_key = serialization.load_pem_private_key(
                private_key_pem, password=passphrase, backend=default_backend()
            )

    def execute_query(
        self, sql: str, statement_count: int, query_tag: str = "", bindings: dict[str, Any] | None = None
    ) -> list[str]:
        """
        Run the query in Snowflake using SnowflakeSQL API by making API request.

        :param sql: the sql string to be executed with possibly multiple statements
        :param statement_count: set the MULTI_STATEMENT_COUNT field to the number of SQL statements
         in the request
        :param query_tag: (Optional) Query tag that you want to associate with the SQL statement.
            For details, see https://docs.snowflake.com/en/sql-reference/parameters.html#label-query-tag
            parameter.
        :param bindings: (Optional) Values of bind variables in the SQL statement.
            When executing the statement, Snowflake replaces placeholders (? and :name) in
            the statement with these specified values.
        """
        self.query_ids = []
        conn_config = self._get_conn_params()

        req_id = uuid.uuid4()
        url = f"{self.account_identifier}.snowflakecomputing.com/api/v2/statements"
        params: dict[str, Any] | None = {"requestId": str(req_id), "async": True, "pageSize": 10}
        headers = self.get_headers()
        sql_is_multi_stmt = ";" in sql.strip()
        if not isinstance(bindings, dict) and bindings is not None:
            raise AirflowException("Bindings should be a dictionary or None.")
        if bindings and sql_is_multi_stmt:
            self.log.warning(
                "Bindings are not supported for multi-statement queries. Bindings will be ignored."
            )
        bindings = bindings or {}
        data = {
            "statement": sql,
            "resultSetMetaData": {"format": "json"},
            # If database, schema, warehouse, role parameters have been provided set them accordingly
            # If either of them has been not (Parent class initializes them to None in that case)
            # set them to what in the Airflow connection configuration
            "database": self.database or conn_config["database"],
            "schema": self.schema or conn_config["schema"],
            "warehouse": self.warehouse or conn_config["warehouse"],
            "role": self.role or conn_config["role"],
            "bindings": bindings,
            "parameters": {
                "MULTI_STATEMENT_COUNT": statement_count,
                "query_tag": query_tag,
            },
        }

        _, json_response = self._make_api_call_with_retries("POST", url, headers, params, data)
        self.log.info("Snowflake SQL POST API response: %s", json_response)
        if "statementHandles" in json_response:
            self.query_ids = json_response["statementHandles"]
        elif "statementHandle" in json_response:
            self.query_ids.append(json_response["statementHandle"])
        else:
            raise AirflowException("No statementHandle/statementHandles present in response")
        return self.query_ids

    def get_headers(self) -> dict[str, Any]:
        """Form auth headers based on either OAuth token or JWT token from private key."""
        conn_config = self._get_conn_params()

        # Use OAuth if refresh_token and client_id and client_secret are provided
        if all(
            [conn_config.get("refresh_token"), conn_config.get("client_id"), conn_config.get("client_secret")]
        ):
            oauth_token = self.get_oauth_token(conn_config=conn_config)
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {oauth_token}",
                "Accept": "application/json",
                "User-Agent": "snowflakeSQLAPI/1.0",
                "X-Snowflake-Authorization-Token-Type": "OAUTH",
            }
            return headers

        # Alternatively, get the JWT token from the connection details and the private key
        if not self.private_key:
            self.get_private_key()

        token = JWTGenerator(
            conn_config["account"],  # type: ignore[arg-type]
            conn_config["user"],  # type: ignore[arg-type]
            private_key=self.private_key,
            lifetime=self.token_life_time,
            renewal_delay=self.token_renewal_delta,
        ).get_token()

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
            "Accept": "application/json",
            "User-Agent": "snowflakeSQLAPI/1.0",
            "X-Snowflake-Authorization-Token-Type": "KEYPAIR_JWT",
        }
        return headers

    def get_oauth_token(
        self,
        conn_config: dict[str, Any] | None = None,
        token_endpoint: str | None = None,
        grant_type: str = "refresh_token",
    ) -> str:
        """Generate temporary OAuth access token using refresh token in connection details."""
        warnings.warn(
            "This method is deprecated. Please use `get_oauth_token` method from `SnowflakeHook` instead. ",
            AirflowProviderDeprecationWarning,
            stacklevel=2,
        )
        return super().get_oauth_token(
            conn_config=conn_config, token_endpoint=token_endpoint, grant_type=grant_type
        )

    def get_request_url_header_params(
        self, query_id: str, url_suffix: str | None = None
    ) -> tuple[dict[str, Any], dict[str, Any], str]:
        """
        Build the request header Url with account name identifier and query id from the connection params.

        :param query_id: statement handles query ids for the individual statements.
        :param url_suffix: Optional path suffix to append to the URL. Must start with '/', e.g. '/cancel' or '/result'.
        """
        req_id = uuid.uuid4()
        header = self.get_headers()
        params = {"requestId": str(req_id)}
        url = f"{self.account_identifier}.snowflakecomputing.com/api/v2/statements/{query_id}"
        if url_suffix:
            url += url_suffix
        return header, params, url

    def check_query_output(self, query_ids: list[str]) -> None:
        """
        Make HTTP request to snowflake SQL API based on the provided query ids and log the response.

        :param query_ids: statement handles query id for the individual statements.
        """
        for query_id in query_ids:
            header, params, url = self.get_request_url_header_params(query_id)
            _, response_json = self._make_api_call_with_retries(
                method="GET", url=url, headers=header, params=params
            )
            self.log.info(response_json)

    def _process_response(self, status_code, resp):
        self.log.info("Snowflake SQL GET statements status API response: %s", resp)
        if status_code == 202:
            return {"status": "running", "message": "Query statements are still running"}
        if status_code == 422:
            error_message = resp.get("message", "Unknown error occurred")
            error_details = []
            if code := resp.get("code"):
                error_details.append(f"Code: {code}")
            if sql_state := resp.get("sqlState"):
                error_details.append(f"SQL State: {sql_state}")
            if statement_handle := resp.get("statementHandle"):
                error_details.append(f"Statement Handle: {statement_handle}")

            if error_details:
                enhanced_message = f"{error_message} ({', '.join(error_details)})"
            else:
                enhanced_message = error_message

            return {"status": "error", "message": enhanced_message}
        if status_code == 200:
            if resp_statement_handles := resp.get("statementHandles"):
                statement_handles = resp_statement_handles
            elif resp_statement_handle := resp.get("statementHandle"):
                statement_handles = [resp_statement_handle]
            else:
                statement_handles = []
            return {
                "status": "success",
                "message": resp["message"],
                "statement_handles": statement_handles,
            }
        return {"status": "error", "message": resp["message"]}

    def get_sql_api_query_status(self, query_id: str) -> dict[str, str | list[str]]:
        """
        Based on the query id async HTTP request is made to snowflake SQL API and return response.

        :param query_id: statement handle id for the individual statements.
        """
        self.log.info("Retrieving status for query id %s", query_id)
        header, params, url = self.get_request_url_header_params(query_id)
        status_code, resp = self._make_api_call_with_retries("GET", url, header, params)
        return self._process_response(status_code, resp)

    def wait_for_query(
        self, query_id: str, raise_error: bool = False, poll_interval: int = 5, timeout: int = 60
    ) -> dict[str, str | list[str]]:
        """
        Wait for query to finish either successfully or with error.

        :param query_id: statement handle id for the individual statement.
        :param raise_error: whether to raise an error if the query failed.
        :param poll_interval: time (in seconds) between checking the query status.
        :param timeout: max time (in seconds) to wait for the query to finish before raising a TimeoutError.

        :raises RuntimeError: If the query status is 'error' and `raise_error` is True.
        :raises TimeoutError: If the query doesn't finish within the specified timeout.
        """
        start_time = time.time()

        while True:
            response = self.get_sql_api_query_status(query_id=query_id)
            self.log.debug("Query status `%s`", response["status"])

            if time.time() - start_time > timeout:
                raise TimeoutError(
                    f"Query `{query_id}` did not finish within the timeout period of {timeout} seconds."
                )

            if response["status"] != "running":
                self.log.info("Query status `%s`", response["status"])
                break

            time.sleep(poll_interval)

        if response["status"] == "error" and raise_error:
            raise RuntimeError(response["message"])

        return response

    def get_result_from_successful_sql_api_query(self, query_id: str) -> list[dict[str, Any]]:
        """
        Based on the query id HTTP requests are made to snowflake SQL API and return result data.

        :param query_id: statement handle id for the individual statement.

        :raises RuntimeError: If the query status is not 'success'.
        """
        self.log.info("Retrieving data for query id %s", query_id)
        header, params, url = self.get_request_url_header_params(query_id)
        status_code, response = self._make_api_call_with_retries("GET", url, header, params)

        if (query_status := self._process_response(status_code, response)["status"]) != "success":
            msg = f"Query must have status `success` to retrieve data; got `{query_status}`."
            raise RuntimeError(msg)

        # Below fields should always be present in response, but added some safety checks
        data = response.get("data", [])
        if not data:
            self.log.warning("No data found in the API response.")
            return []
        metadata = response.get("resultSetMetaData", {})
        col_names = [row["name"] for row in metadata.get("rowType", [])]
        if not col_names:
            self.log.warning("No column metadata found in the API response.")
            return []

        num_partitions = len(metadata.get("partitionInfo", []))
        if num_partitions > 1:
            self.log.debug("Result data is returned as multiple partitions. Will perform additional queries.")
            url += "?partition="
            for partition_no in range(1, num_partitions):  # First partition was already returned
                self.log.debug("Querying for partition no. %s", partition_no)
                _, response = self._make_api_call_with_retries("GET", url + str(partition_no), header, params)
                data.extend(response.get("data", []))

        return [dict(zip(col_names, row)) for row in data]  # Merged column names with data

    async def get_sql_api_query_status_async(self, query_id: str) -> dict[str, str | list[str]]:
        """
        Based on the query id async HTTP request is made to snowflake SQL API and return response.

        :param query_id: statement handle id for the individual statements.
        """
        self.log.info("Retrieving status for query id %s", query_id)
        header, params, url = self.get_request_url_header_params(query_id)
        status_code, resp = await self._make_api_call_with_retries_async("GET", url, header, params)
        return self._process_response(status_code, resp)

    def _cancel_sql_api_query_execution(self, query_id: str) -> dict[str, str | list[str]]:
        self.log.info("Cancelling query id %s", query_id)
        header, params, url = self.get_request_url_header_params(query_id, "/cancel")
        status_code, resp = self._make_api_call_with_retries("POST", url, header, params)
        return self._process_response(status_code, resp)

    def cancel_queries(self, query_ids: list[str]) -> None:
        for query_id in query_ids:
            self._cancel_sql_api_query_execution(query_id)

    @staticmethod
    def _should_retry_on_error(exception) -> bool:
        """
        Determine if the exception should trigger a retry based on error type and status code.

        Retries on HTTP errors 429 (Too Many Requests), 503 (Service Unavailable),
        and 504 (Gateway Timeout) as recommended by Snowflake error handling docs.
        Retries on connection errors and timeouts.

        :param exception: The exception to check
        :return: True if the request should be retried, False otherwise
        """
        if isinstance(exception, HTTPError):
            return exception.response.status_code in [429, 503, 504]
        if isinstance(exception, ClientResponseError):
            return exception.status in [429, 503, 504]
        if isinstance(
            exception,
            ConnectionError | Timeout | ClientConnectionError,
        ):
            return True
        return False

    def _make_api_call_with_retries(
        self, method: str, url: str, headers: dict, params: dict | None = None, json: dict | None = None
    ):
        """
        Make an API call to the Snowflake SQL API with retry logic for specific HTTP errors.

        Error handling implemented based on Snowflake error handling docs:
        https://docs.snowflake.com/en/developer-guide/sql-api/handling-errors

        :param method: The HTTP method to use for the API call.
        :param url: The URL for the API endpoint.
        :param headers: The headers to include in the API call.
        :param params: (Optional) The query parameters to include in the API call.
        :param json: (Optional) The data to include in the API call.
        :return: The response object from the API call.
        """
        with requests.Session() as session:
            for attempt in Retrying(**self.retry_config):  # type: ignore
                with attempt:
                    if method.upper() in ("GET", "POST"):
                        response = session.request(
                            method=method.lower(), url=url, headers=headers, params=params, json=json
                        )
                    else:
                        raise ValueError(f"Unsupported HTTP method: {method}")
                    response.raise_for_status()
                    return response.status_code, response.json()

    async def _make_api_call_with_retries_async(self, method, url, headers, params=None):
        """
        Make an API call to the Snowflake SQL API asynchronously with retry logic for specific HTTP errors.

        Error handling implemented based on Snowflake error handling docs:
        https://docs.snowflake.com/en/developer-guide/sql-api/handling-errors

        :param method: The HTTP method to use for the API call. Only GET is supported as  is synchronous.
        :param url: The URL for the API endpoint.
        :param headers: The headers to include in the API call.
        :param params: (Optional) The query parameters to include in the API call.
        :return: The response object from the API call.
        """
        async with aiohttp.ClientSession(headers=headers) as session:
            async for attempt in AsyncRetrying(**self.retry_config):
                with attempt:
                    if method.upper() == "GET":
                        async with session.request(method=method.lower(), url=url, params=params) as response:
                            response.raise_for_status()
                            # Return status and json content for async processing
                            content = await response.json()
                            return response.status, content
                    else:
                        raise ValueError(f"Unsupported HTTP method: {method}")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Abstract operator that child classes implement ``COPY INTO <TABLE> SQL in Snowflake``."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from airflow.providers.common.compat.sdk import BaseOperator
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook
from airflow.providers.snowflake.utils.common import enclose_param


def _validate_parameter(param_name: str, value: str | None) -> str | None:
    """Validate that the parameter doesn't contain any invalid pattern."""
    if value is None:
        return None
    if ";" in value:
        raise ValueError(f"Invalid {param_name}: semicolons (;) not allowed.")
    return value


class CopyFromExternalStageToSnowflakeOperator(BaseOperator):
    """
    Executes a COPY INTO command to load files from an external stage from clouds to Snowflake.

    This operator requires the snowflake_conn_id connection. The snowflake host, login,
    and, password field must be setup in the connection. Other inputs can be defined
    in the connection or hook instantiation.

    :param namespace: snowflake namespace
    :param table: snowflake table
    :param file_format: file format name i.e. CSV, AVRO, etc
    :param stage: reference to a specific snowflake stage. If the stage's schema is not the same as the
        table one, it must be specified
    :param prefix: cloud storage location specified to limit the set of files to load
    :param files: files to load into table
    :param pattern: pattern to load files from external location to table
    :param snowflake_conn_id:  Reference to :ref:`Snowflake connection id<howto/connection:snowflake>`
    :param account: snowflake account name
    :param warehouse: name of snowflake warehouse
    :param database: name of snowflake database
    :param region: name of snowflake region
    :param role: name of snowflake role
    :param schema: name of snowflake schema
    :param authenticator: authenticator for Snowflake.
        'snowflake' (default) to use the internal Snowflake authenticator
        'externalbrowser' to authenticate using your web browser and
        Okta, ADFS or any other SAML 2.0-compliant identify provider
        (IdP) that has been defined for your account
        ``https://<your_okta_account_name>.okta.com`` to authenticate
        through native Okta.
    :param session_parameters: You can set session-level parameters at
        the time you connect to Snowflake
    :param copy_options: snowflake COPY INTO syntax copy options
    :param validation_mode: snowflake COPY INTO syntax validation mode

    """

    template_fields: Sequence[str] = ("files",)
    template_fields_renderers = {"files": "json"}

    def __init__(
        self,
        *,
        files: list | None = None,
        table: str,
        stage: str,
        prefix: str | None = None,
        file_format: str,
        schema: str | None = None,
        columns_array: list | None = None,
        pattern: str | None = None,
        warehouse: str | None = None,
        database: str | None = None,
        autocommit: bool = True,
        snowflake_conn_id: str = "snowflake_default",
        role: str | None = None,
        authenticator: str | None = None,
        session_parameters: dict | None = None,
        copy_options: str | None = None,
        validation_mode: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.files = files
        self.table = _validate_parameter("table", table)
        self.stage = _validate_parameter("stage", stage)
        self.prefix = prefix
        self.file_format = file_format
        self.schema = schema
        self.columns_array = columns_array
        self.pattern = pattern
        self.warehouse = warehouse
        self.database = database
        self.autocommit = autocommit
        self.snowflake_conn_id = snowflake_conn_id
        self.role = role
        self.authenticator = authenticator
        self.session_parameters = session_parameters
        self.copy_options = copy_options
        self.validation_mode = validation_mode

        self.hook: SnowflakeHook | None = None
        self._sql: str | None = None
        self._result: list[dict[str, Any]] = []

    def execute(self, context: Any) -> None:
        self.hook = SnowflakeHook(
            snowflake_conn_id=self.snowflake_conn_id,
            warehouse=self.warehouse,
            database=self.database,
            role=self.role,
            schema=self.schema,
            authenticator=self.authenticator,
            session_parameters=self.session_parameters,
        )

        if self.schema:
            into = f"{self.schema}.{self.table}"
        else:
            into = self.table  # type: ignore[assignment]

        if self.columns_array:
            into = f"{into}({', '.join(self.columns_array)})"

        self._sql = f"""
        COPY INTO {into}
             FROM  @{self.stage}/{self.prefix or ""}
        {"FILES=(" + ",".join(map(enclose_param, self.files)) + ")" if self.files else ""}
        {"PATTERN=" + enclose_param(self.pattern) if self.pattern else ""}
        FILE_FORMAT={self.file_format}
        {self.copy_options or ""}
        {self.validation_mode or ""}
        """
        self.log.info("Executing COPY command...")
        self._result = self.hook.run(  # type: ignore # mypy does not work well with return_dictionaries=True
            sql=self._sql,
            autocommit=self.autocommit,
            handler=lambda x: x.fetchall(),
            return_dictionaries=True,
        )
        self.log.info("COPY command completed")

    @staticmethod
    def _extract_openlineage_unique_dataset_paths(
        query_result: list[dict[str, Any]],
    ) -> tuple[list[tuple[str, str]], list[str]]:
        """
        Extract and return unique OpenLineage dataset paths and file paths that failed to be parsed.

        Each row in the results is expected to have a 'file' field, which is a URI.
        The function parses these URIs and constructs a set of unique OpenLineage (namespace, name) tuples.
        Additionally, it captures any URIs that cannot be parsed or processed
        and returns them in a separate error list.

        For Azure, Snowflake has a unique way of representing URI:
            azure://<account_name>.blob.core.windows.net/<container_name>/path/to/file.csv
        that is transformed by this function to a Dataset with more universal naming convention:
            Dataset(namespace="wasbs://container_name@account_name", name="path/to"), as described at
        https://github.com/OpenLineage/OpenLineage/blob/main/spec/Naming.md#wasbs-azure-blob-storage

        :param query_result: A list of dictionaries, each containing a 'file' key with a URI value.
        :return: Two lists - the first is a sorted list of tuples, each representing a unique dataset path,
         and the second contains any URIs that cannot be parsed or processed correctly.

        >>> method = CopyFromExternalStageToSnowflakeOperator._extract_openlineage_unique_dataset_paths

        >>> results = [{"file": "azure://my_account.blob.core.windows.net/azure_container/dir3/file.csv"}]
        >>> method(results)
        ([('wasbs://azure_container@my_account', 'dir3/file.csv')], [])

        >>> results = [{"file": "azure://my_account.blob.core.windows.net/azure_container"}]
        >>> method(results)
        ([('wasbs://azure_container@my_account', '/')], [])

        >>> results = [{"file": "s3://bucket"}, {"file": "gcs://bucket/"}, {"file": "s3://bucket/a.csv"}]
        >>> method(results)
        ([('gcs://bucket', '/'), ('s3://bucket', '/'), ('s3://bucket', 'a.csv')], [])

        >>> results = [{"file": "s3://bucket/dir/file.csv"}, {"file": "gcs://bucket/dir/dir2/a.txt"}]
        >>> method(results)
        ([('gcs://bucket', 'dir/dir2/a.txt'), ('s3://bucket', 'dir/file.csv')], [])

        >>> results = [
        ...     {"file": "s3://bucket/dir/file.csv"},
        ...     {"file": "azure://my_account.something_new.windows.net/azure_container"},
        ... ]
        >>> method(results)
        ([('s3://bucket', 'dir/file.csv')], ['azure://my_account.something_new.windows.net/azure_container'])

        >>> results = [
        ...     {"file": "s3://bucket/dir/file.csv"},
        ...     {"file": "s3:/invalid-s3-uri"},
        ...     {"file": "gcs:invalid-gcs-uri"},
        ... ]
        >>> method(results)
        ([('s3://bucket', 'dir/file.csv')], ['gcs:invalid-gcs-uri', 's3:/invalid-s3-uri'])
        """
        import re
        from urllib.parse import urlparse

        azure_regex = r"azure:\/\/(\w+)?\.blob.core.windows.net\/(\w+)\/?(.*)?"
        extraction_error_files = []
        unique_dataset_paths = set()

        for row in query_result:
            try:
                uri = urlparse(row["file"])

                # Check for valid URI structure
                if not uri.scheme or not uri.netloc:
                    extraction_error_files.append(row["file"])
                    continue

                if uri.scheme == "azure":
                    match = re.fullmatch(azure_regex, row["file"])
                    if not match:
                        extraction_error_files.append(row["file"])
                        continue
                    account_name, container_name, name = match.groups()
                    namespace = f"wasbs://{container_name}@{account_name}"
                else:
                    namespace = f"{uri.scheme}://{uri.netloc}"
                    name = uri.path.lstrip("/")

                if name in ("", "."):
                    name = "/"

                unique_dataset_paths.add((namespace, name))
            except Exception:
                extraction_error_files.append(row["file"])

        return sorted(unique_dataset_paths), sorted(extraction_error_files)

    def get_openlineage_facets_on_complete(self, task_instance):
        """Implement _on_complete because we rely on return value of a query."""
        import re

        from airflow.providers.common.compat.openlineage.facet import (
            Dataset,
            Error,
            ExternalQueryRunFacet,
            ExtractionErrorRunFacet,
            SQLJobFacet,
        )
        from airflow.providers.openlineage.extractors import OperatorLineage
        from airflow.providers.openlineage.sqlparser import SQLParser

        if not self._sql:
            return OperatorLineage()

        query_results = self._result or []

        # This typically happens when no files were processed (empty directory)
        if len(query_results) == 1 and ("file" not in query_results[0] or query_results[0]["file"] is None):
            query_results = []

        unique_dataset_paths, extraction_error_files = self._extract_openlineage_unique_dataset_paths(
            query_results
        )
        input_datasets = [Dataset(namespace=namespace, name=name) for namespace, name in unique_dataset_paths]

        run_facets = {}
        if extraction_error_files:
            self.log.debug(
                "Unable to extract Dataset namespace and name for the following files: `%s`.",
                extraction_error_files,
            )
            run_facets["extractionError"] = ExtractionErrorRunFacet(
                totalTasks=len(query_results),
                failedTasks=len(extraction_error_files),
                errors=[
                    Error(
                        errorMessage="Unable to extract Dataset namespace and name.",
                        stackTrace=None,
                        task=file_uri,
                        taskNumber=None,
                    )
                    for file_uri in extraction_error_files
                ],
            )

        connection = self.hook.get_connection(getattr(self.hook, str(self.hook.conn_name_attr)))
        database_info = self.hook.get_openlineage_database_info(connection)

        dest_name = self.table
        schema = self.hook.get_openlineage_default_schema()
        database = database_info.database
        if schema:
            dest_name = f"{schema}.{dest_name}"
            if database:
                dest_name = f"{database}.{dest_name}"

        snowflake_namespace = SQLParser.create_namespace(database_info)
        query = SQLParser.normalize_sql(self._sql)
        query = re.sub(r"\n+", "\n", re.sub(r" +", " ", query))

        run_facets["externalQuery"] = ExternalQueryRunFacet(
            externalQueryId=self.hook.query_ids[0], source=snowflake_namespace
        )

        return OperatorLineage(
            inputs=input_datasets,
            outputs=[Dataset(namespace=snowflake_namespace, name=dest_name)],
            job_facets={"sql": SQLJobFacet(query=query)},
            run_facets=run_facets,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#
from __future__ import annotations

from unittest.mock import MagicMock

import pytest
from snowflake.connector import DictCursor

from airflow.providers.common.sql.hooks.handlers import fetch_all_handler
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook

TASK_ID = "sql-operator"
HOST = "host"
DEFAULT_CONN_ID = "snowflake_default"
PASSWORD = "password"


class SnowflakeHookForTests(SnowflakeHook):
    conn_name_attr = "snowflake_conn_id"
    get_conn = MagicMock(name="conn")


def get_cursor_descriptions(fields: list[str]) -> list[tuple[str]]:
    return [(field,) for field in fields]


@pytest.mark.parametrize(
    (
        "return_last",
        "split_statements",
        "sql",
        "cursor_calls",
        "cursor_descriptions",
        "cursor_results",
        "hook_descriptions",
        "hook_results",
        "return_dictionaries",
    ),
    [
        pytest.param(
            True,
            False,
            "select * from test.test",
            ["select * from test.test"],
            [["id", "value"]],
            ([[1, 2], [11, 12]],),
            [[("id",), ("value",)]],
            [[1, 2], [11, 12]],
            False,
            id="The return_last set and no split statements set on single query in string",
        ),
        pytest.param(
            False,
            False,
            "select * from test.test;",
            ["select * from test.test;"],
            [["id", "value"]],
            ([[1, 2], [11, 12]],),
            [[("id",), ("value",)]],
            [[1, 2], [11, 12]],
            False,
            id="The return_last not set and no split statements set on single query in string",
        ),
        pytest.param(
            True,
            True,
            "select * from test.test;",
            ["select * from test.test;"],
            [["id", "value"]],
            ([[1, 2], [11, 12]],),
            [[("id",), ("value",)]],
            [[1, 2], [11, 12]],
            False,
            id="The return_last set and split statements set on single query in string",
        ),
        pytest.param(
            False,
            True,
            "select * from test.test;",
            ["select * from test.test;"],
            [["id", "value"]],
            ([[1, 2], [11, 12]],),
            [[("id",), ("value",)]],
            [[[1, 2], [11, 12]]],
            False,
            id="The return_last not set and split statements set on single query in string",
        ),
        pytest.param(
            True,
            True,
            "select * from test.test;select * from test.test2;",
            ["select * from test.test;", "select * from test.test2;"],
            [["id", "value"], ["id2", "value2"]],
            ([[1, 2], [11, 12]], [[3, 4], [13, 14]]),
            [[("id2",), ("value2",)]],
            [[3, 4], [13, 14]],
            False,
            id="The return_last set and split statements set on multiple queries in string",
        ),  # Failing
        pytest.param(
            False,
            True,
            "select * from test.test;select * from test.test2;",
            ["select * from test.test;", "select * from test.test2;"],
            [["id", "value"], ["id2", "value2"]],
            ([[1, 2], [11, 12]], [[3, 4], [13, 14]]),
            [[("id",), ("value",)], [("id2",), ("value2",)]],
            [[[1, 2], [11, 12]], [[3, 4], [13, 14]]],
            False,
            id="The return_last not set and split statements set on multiple queries in string",
        ),
        pytest.param(
            True,
            True,
            ["select * from test.test;"],
            ["select * from test.test"],
            [["id", "value"]],
            ([[1, 2], [11, 12]],),
            [[("id",), ("value",)]],
            [[[1, 2], [11, 12]]],
            False,
            id="The return_last set on single query in list",
        ),
        pytest.param(
            False,
            True,
            ["select * from test.test;"],
            ["select * from test.test"],
            [["id", "value"]],
            ([[1, 2], [11, 12]],),
            [[("id",), ("value",)]],
            [[[1, 2], [11, 12]]],
            False,
            id="The return_last not set on single query in list",
        ),
        pytest.param(
            True,
            True,
            "select * from test.test;select * from test.test2;",
            ["select * from test.test", "select * from test.test2"],
            [["id", "value"], ["id2", "value2"]],
            ([[1, 2], [11, 12]], [[3, 4], [13, 14]]),
            [[("id2",), ("value2",)]],
            [[3, 4], [13, 14]],
            False,
            id="The return_last set on multiple queries in list",
        ),
        pytest.param(
            False,
            True,
            "select * from test.test;select * from test.test2;",
            ["select * from test.test", "select * from test.test2"],
            [["id", "value"], ["id2", "value2"]],
            ([[1, 2], [11, 12]], [[3, 4], [13, 14]]),
            [[("id",), ("value",)], [("id2",), ("value2",)]],
            [[[1, 2], [11, 12]], [[3, 4], [13, 14]]],
            False,
            id="The return_last not set on multiple queries not set",
        ),
        pytest.param(
            False,
            True,
            "select * from test.test;select * from test.test2;",
            ["select * from test.test", "select * from test.test2"],
            [["id", "value"], ["id2", "value2"]],
            (
                [{"id": 1, "value": 2}, {"id": 11, "value": 12}],
                [{"id2": 3, "value2": 4}, {"id2": 13, "value2": 14}],
            ),
            [[("id",), ("value",)], [("id2",), ("value2",)]],
            [
                [{"id": 1, "value": 2}, {"id": 11, "value": 12}],
                [{"id2": 3, "value2": 4}, {"id2": 13, "value2": 14}],
            ],
            True,
            id="DictCursor: The return_last not set on multiple queries not set",
        ),
    ],
)
def test_query(
    return_last,
    split_statements,
    sql,
    cursor_calls,
    cursor_descriptions,
    cursor_results,
    hook_descriptions,
    hook_results,
    return_dictionaries,
):
    modified_descriptions = [
        get_cursor_descriptions(cursor_description) for cursor_description in cursor_descriptions
    ]
    dbapi_hook = SnowflakeHookForTests()
    dbapi_hook.get_conn.return_value.cursor.return_value.rowcount = 2
    dbapi_hook.get_conn.return_value.cursor.return_value._description_index = 0

    def mock_execute(*args, **kwargs):
        # the run method accesses description property directly, and we need to modify it after
        # every execute, to make sure that different descriptions are returned. I could not find easier
        # method with mocking
        dbapi_hook.get_conn.return_value.cursor.return_value.description = modified_descriptions[
            dbapi_hook.get_conn.return_value.cursor.return_value._description_index
        ]
        dbapi_hook.get_conn.return_value.cursor.return_value._description_index += 1

    dbapi_hook.get_conn.return_value.cursor.return_value.execute = mock_execute
    dbapi_hook.get_conn.return_value.cursor.return_value.fetchall.side_effect = cursor_results
    results = dbapi_hook.run(
        sql=sql,
        handler=fetch_all_handler,
        return_last=return_last,
        split_statements=split_statements,
        return_dictionaries=return_dictionaries,
    )

    assert dbapi_hook.descriptions == hook_descriptions
    assert dbapi_hook.last_description == hook_descriptions[-1]
    assert results == hook_results

    if return_dictionaries:
        dbapi_hook.get_conn.return_value.cursor.assert_called_with(DictCursor)
    else:
        dbapi_hook.get_conn.return_value.cursor.assert_called_with()
    dbapi_hook.get_conn.return_value.cursor.return_value.close.assert_called()


@pytest.mark.parametrize(
    "empty_statement",
    [
        pytest.param([], id="Empty list"),
        pytest.param("", id="Empty string"),
        pytest.param("\n", id="Only EOL"),
    ],
)
def test_no_query(empty_statement):
    dbapi_hook = SnowflakeHookForTests()
    dbapi_hook.get_conn.return_value.cursor.rowcount = 0
    with pytest.raises(ValueError, match="List of SQL statements is empty"):
        dbapi_hook.run(sql=empty_statement)
